# -------------------------------------------------------------------
#
#	sjtCopySkinWeights v1.0
#
#	author: Sveinbjorn J. Tryggvason <sjt@sjt.is>
#
#	== USAGE ==
#
#	1. Select the object with the correct weights.
#	2. Select the object with the incorrect weights.
#	3. Run the command e.g. 'sjtCopySkinWeights -s "L_" -r "R_"'
#
#	== ABOUT ==
#
#	A maya command plugin with the command 'sjtCopySkinWeights' to copy
#	weights from one object to another with some rather significant restrictions.
#		1. The meshes must be exactly mirrors of each other vertex-number wise. 
#		   This plugin simply uses the vertex numbers of the mesh so each vertex
#		   on object A must have a corresponding vertex on object B.
#		2. The joint names must be somehow side-differentiated e.g. L_joint and R_joint or
#		   left_side_arm_JNT and right_side_arm_JNT or you must somehow be able to map one side of 
#		   joint names to the other.
#	
#	This command accepts two arguments -search/-s and -replace/-r and each argument requres
#	a string to follow it that tells the plugin how to map the right joint names to the
#	left ones. Yes it works kind of backwards. The command first constructs a list of all the
#	influences on the first object, then it goes over every vertex on the other object and tries
#	to map it's influences to the ones in the influence-list from the first object.
#	
#	Example:
#		one one side the joints are called 'L_....._JNT' and the other 'R_......_JNT' the proper
#		arguments for this command to work would be (MEL syntax):
#				
#			sjtCopySkinWeights -s "L_" -r "R_" (if no arguments are givien, this is the default search pattern)
#
#		this would also have worked (since the search strings can be regular expressions):
#		
#			sjtCopySkinWeights -s "L_" -r "^[rR]_"
#		
#		This would allow the right side joints to be named either 'R_...._JNT' or 'r_...._JNT'
#		Just remember the regular expression should go with the -r flag, the -s flag is just used to replace
#		The match from the -r flag with.
#
#	== EXAMPLE FILE ==
#	
#	To test the command immediately you can load up the included test file 'copyWeights_example.ma'. First 
#	try scrubbing in the timeline to see the difference in weighting. The right side (pCube1) bends in the middle 
#	but the left side(pCube2) does not. To copy the weights from pCube1 to pCube2 first load the plugin, 
#	select pCube1 then pCube2 and type (MEL syntax):
#		sjtCopySkinWeights
#	into the command line (or script editor) and then the weights will have been copied from pCube1 to pCube2. Voila.
#
#	== INSTALLATION ==
#
#	1. Copy this file to any folder on your MAYA_PLUG_IN_PATH.
#	2. Start Maya
#	3. Enjoy
#
# -------------------------------------------------------------------


import sys, re
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.OpenMayaAnim as omA

kPluginCmdName = 'sjtCopySkinWeights'

kSearchFlag = "-s"
kSearchLongFlag = "-search"
kReplaceFlag = "-r"
kReplaceLongFlag = "-right"

class sjtCopySkinWeights(omMPx.MPxCommand):
	def __init__(self):
		omMPx.MPxCommand.__init__(self)
		self.mapping = ['L_','R_']
	
	def doIt(self,argList):
		# set up our variables
		self.node = om.MObject()		# the selected object (transform node)
		self.nodePath = om.MDagPath()	# the dag path to the selected object
		self.list = om.MSelectionList()	# the selection list
		
		
		# parse the command arguments
		argData = om.MArgDatabase(self.syntax(), argList)
		if argData.isFlagSet(kSearchFlag):
			self.mapping[0] = argData.flagArgumentString(kSearchFlag, 0)
		if argData.isFlagSet(kReplaceFlag):
			self.mapping[1] = argData.flagArgumentString(kReplaceFlag, 0)
			
		
		om.MGlobal.getActiveSelectionList(self.list) # get the selection list
		if self.list.length()!=2:
			om.MGlobal.displayError('[%s] wrong number of objects selected, first select source object, then destination object!' % kPluginCmdName)
			return
		self.list.getDependNode(0,self.node)		 # get the MObject for the source object
		
		# find the connected skinCluster, this will be our source skin object
		skinCluster = self.findSkinCluster(self.node)
		# get a dag path to the selected source object
		om.MDagPath.getAPathTo(self.node,self.nodePath)
		
		# create a skinCluster function set
		skin = omA.MFnSkinCluster(skinCluster)
		
		
		# create an integer 'pointer' to use with the getWeights() command
		integer = om.MScriptUtil()
		integerPointer = integer.asUintPtr()
		self.integerPointer = integerPointer
		
		# lets note all the influence objects influencing the source skin
		influenceObjects = om.MDagPathArray()
		skin.influenceObjects(influenceObjects)
		
		# we use this array to store our weight values, each line is a vertex and each line consists of a dictionary of the form:
		# [influenceObject]->weight,[influenceObject]->weight,...
		self.weightMap = []
		self.influenceArray = []
		# note the short names of all the influence objects
		for i in range(influenceObjects.length()):
			self.influenceArray.append( influenceObjects[i].partialPathName().split('|')[-1] )
		
		# iterate over all the vertexes of the selected source object
		iterator = om.MItMeshVertex(self.node)
		while not iterator.isDone():
			# the current vertex
			vert =  iterator.currentItem()
			# an array to store the weights for all the influence objects for this vertex
			array = om.MDoubleArray()
			# get the weights
			skin.getWeights(self.nodePath,vert,array,integerPointer)
			
			# just a check for a state that _should_ never arise
			if len(array)!=len(self.influenceArray):
				raise RuntimeError("Number of influences and weight values does not match!")
			
			# the weight dictionary
			vertDict = {}
			for weight,influence in zip(array,self.influenceArray):
				vertDict[influence] = weight
			
			# add this to our master weight map array
			self.weightMap.append(vertDict)
			# go to the next vertex
			iterator.next()
			
		# in the next step we apply the weights to the target object
		self.redoIt()
	
	def redoIt(self):
		if self.list.length()!=2:
			om.MGlobal.displayError('[%s] wrong number of objects selected, first select source object, then destination object!' % kPluginCmdName)
			return
		# set up our objects
		self.targetNode = om.MObject() # the target object (transform node)
		
		self.list.getDependNode(1,self.targetNode) # fetch the MObject from the master selection list
		# and fetch a dag path to the target object as well
		self.targetNodePath = om.MDagPath()
		om.MDagPath.getAPathTo(self.targetNode,self.targetNodePath)
		
		# find the connected skinCluster
		self.targetSkinCluster = self.findSkinCluster(self.targetNode)
		# and create a function set for that cluster
		self.targetSkin = omA.MFnSkinCluster(self.targetSkinCluster)
		
		# we want to note all the influences influencing the target skin
		targetInfluences = om.MDagPathArray()
		self.targetSkin.influenceObjects(targetInfluences)
		
		# the influenceIndices is simply an array with the index of each of the influences
		self.targetInfluenceNames = []
		self.influenceIndices = om.MIntArray()
		for i in range(targetInfluences.length()):
			self.targetInfluenceNames.append(targetInfluences[i].partialPathName().split('|')[-1])
			self.influenceIndices.append(i)
		
		# iterate over all the verticies of the target object
		tIterator = om.MItMeshVertex(self.targetNode)
		# we want to keep the original weights around so we can undo
		self.originalWeightMap = []
		index = 0
		while not tIterator.isDone():
			tVert = tIterator.currentItem() # the current vertex
			originalWeights = om.MDoubleArray()	# make sure the originalWeights array is empty
			# get the weighs to use for undo
			self.targetSkin.getWeights(self.targetNodePath,tVert,originalWeights,self.integerPointer)
			# and store them in the appropriate array
			self.originalWeightMap.append(originalWeights)
			
			# now we iterate over the influences influencing the target skin and apply the weights for each influence seperately
			for tInfluence in self.targetInfluenceNames:
				tmp = om.MDoubleArray()
				# convert the names of the influences with our search/replace parameters
				sourceInfluence = re.sub(self.mapping[1], self.mapping[0], tInfluence)#tInfluence.replace(self.mapping[1],self.mapping[0])
				# get the index of the target influence
				sourceInfluenceIndex = self.targetInfluenceNames.index(tInfluence)
				# set the weights for the given influence for the current vertex
				try:
					self.targetSkin.setWeights(self.targetNodePath,tVert,sourceInfluenceIndex,self.weightMap[index][sourceInfluence],False,tmp)
				except KeyError:
					om.MGlobal.displayWarning('Unable to find influence %s for vertex %d' % ( tInfluence, tIterator.index()))
					continue
			
			index = index+1
			tIterator.next()
		
	
	def undoIt(self):
		# undo
		tIterator = om.MItMeshVertex(self.targetNode)
		index = 0
		while not tIterator.isDone():
			tVert = tIterator.currentItem()
			# apply the weights we stored in redoIt
			self.targetSkin.setWeights(self.targetNodePath,tVert,self.influenceIndices,self.originalWeightMap[index],False,om.MDoubleArray())
			index = index+1
			tIterator.next()
	
	def isUndoable(self):
		return True
	
	def findSkinCluster(self,node):
		'''
		A handy method that traverses the dag to find a skincluster upstream.
		'''
		mfndag = om.MFnDagNode()
		# check to see if the passed node is a transform node. If not we error out.
		if node.apiType() != om.MFn.kTransform:
			om.MGlobal.displayError("Selected object is not a transform node.")
			return None
		# attach the function set to the node
		mfndag.setObject(node)
		
		# find the child, i.e. shape node
		try:
			theShapeNode = mfndag.child(0)
		except:
			om.MGlobal.displayError("Unable to find corresponding shape node.")
			return None
		# make sure that found child is a shape node
		if not theShapeNode.hasFn(om.MFn.kMesh):
			om.MGlobal.displayError("Unable to find corresponding shape node.")
			return None
		
		# iterate over the dependency graph to find the attached skinCluster,
		# if none is found we error out.
		try:
			iterator = om.MItDependencyGraph(	theShapeNode,
												om.MFn.kSkinClusterFilter,
												om.MItDependencyGraph.kUpstream,
												om.MItDependencyGraph.kBreadthFirst,
												om.MItDependencyGraph.kNodeLevel)
		except:
			om.MGlobal.displayError('No skin cluster found.')
			return None
		
		# return the found skinCluster
		return iterator.currentItem()

def cmdCreator():
	return omMPx.asMPxPtr( sjtCopySkinWeights() )

def syntaxCreator():
	syntax = om.MSyntax()
	syntax.addFlag(kSearchFlag, kSearchLongFlag, om.MSyntax.kString)
	syntax.addFlag(kReplaceFlag, kReplaceLongFlag, om.MSyntax.kString)
	return syntax

def initializePlugin(mobject):
	mplugin = omMPx.MFnPlugin(mobject)
	try:
		mplugin.registerCommand( kPluginCmdName, cmdCreator, syntaxCreator )
	except:
		sys.stderr.write("Failed to register command: %s\n"% kPluginCmdName)
		raise

def uninitializePlugin(mobject):
	mplugin = omMPx.MFnPlugin(mobject)
	try:
		mplugin.deregisterCommand( kPluginCmdName )
	except:
		sys.stderr.write("Failed to unregister command: %s\n" % kPluginCmdName )
		raise
